/*
 * This file was automatically generated from zlib.h,
 * which is covered by the following license:
 * Copyright (C) 1995-2022 Jean-loup Gailly and Mark Adler
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 * Jean-loup Gailly        Mark Adler
 * jloup@gzip.org          madler@alumni.caltech.edu
 */
FUNC(int, compress, (unsigned char *dest, unsigned long *destLen, const unsigned char *source, unsigned long sourceLen), (dest, destLen, source, sourceLen))
FUNC(unsigned long, compressBound, (unsigned long sourceLen), (sourceLen))
FUNC(int, uncompress, (unsigned char *dest, unsigned long *destLen, const unsigned char *source, unsigned long sourceLen), (dest, destLen, source, sourceLen))
